#include "gaussLU.hh"
#include "recursiveLU.hh"
#include "alglin.hh"
#include "timing.hh"
#include <iostream>

using namespace std ;
using gaussLU::indexType ;
using gaussLU::valueType ;

int
main(int argc, char * const argv[]) {

  Timing tm ;
  valueType * A[3] ;
  valueType * b[3] ;
  
  indexType N = 1000 ;
  A[0] = new valueType[ N * N ] ;
  A[1] = new valueType[ N * N ] ;
  A[2] = new valueType[ N * N ] ;
  b[0] = new valueType[ N ] ;
  b[1] = new valueType[ N ] ;
  b[2] = new valueType[ N ] ;

  valueType * x    = new valueType[ N ] ;
  indexType * perm = new indexType[ N ] ;
  int       * IPIV = new int[ N ] ;
  valueType * work = new valueType[ N ] ;

  valueType LUelapsed[3] ;
  valueType SOLelapsed[3] ;
  valueType ERR[3] ;

  for ( indexType nn=0 ; nn < 10 ; ++nn ) {

    indexType lda = N ;

    // fill the matrix randomly   
    for ( indexType i = 0 ; i < N ; ++i )
      for ( indexType j = 0 ; j < N ; ++j )
        A[0][i+j*lda] = arc4random() * 1.02323E-8 ;
    // fill the solution vector 
    for ( indexType i = 0 ; i < N ; ++i ) x[i] = i ;

    // multipliy
    alglin::gemm( "N", "N",
                  N, 1, N, 
                  1.0, A[0], lda, 
                       x, lda,
                  0.0, b[0], lda ) ;

    alglin::copy( N*N, A[0], 1, A[1], 1 ) ;
    alglin::copy( N*N, A[0], 1, A[2], 1 ) ;
    alglin::copy( N,   b[0], 1, b[1], 1 ) ;
    alglin::copy( N,   b[0], 1, b[2], 1 ) ;

    tm . start() ;
    valueType cond = gaussLU::factorize( N, A[0], lda, perm, work ) ;
    LUelapsed[0] = tm . milliseconds() ;

    tm . start() ;
    gaussLU::solve( N, A[0], lda, perm, b[0] ) ;
    //gaussLU::solveTranspose( N, A[0], lda, perm, b[0] ) ;
    SOLelapsed[0] = tm . milliseconds() ;

    alglin::axpy( N, -1, x, 1,  b[0], 1 ) ;
    ERR[0] = alglin::asum( N, b[0], 1 ) ;

    ///////////////////////////

    tm . start() ;
    //cond = gaussLU::factorBlock( N, A[1], lda, perm, work ) ;
    recursiveLU::factorize( N, N, A[1], lda, perm ) ;
    LUelapsed[1] = tm . milliseconds() ;

    tm . start() ;
    alglin::getrs( "N", N, 1, A[1], lda, (int*)perm, b[1], lda ) ; 
    //gaussLU::solve( N, A[1], lda, perm, b[1] ) ;
    //gaussLU::solveTranspose( N, A[1], lda, perm, b[1] ) ;
    SOLelapsed[1] = tm . milliseconds() ;

    alglin::axpy( N, -1, x, 1,  b[1], 1 ) ;
    ERR[1] = alglin::asum( N, b[1], 1 ) ;

    /////////////////////////////

    tm . start() ;
    alglin::getrf( N, N, A[2], lda, IPIV ) ;
    LUelapsed[2] = tm . milliseconds() ;

    tm . start() ;
    alglin::getrs( "N", N, 1, A[2], lda, IPIV, b[2], lda ) ; 
    SOLelapsed[2] = tm . milliseconds() ;

    alglin::axpy( N, -1, x, 1, b[2], 1 ) ;
    ERR[2] = alglin::asum( N, b[2], 1 ) ;

    /////////////////////////////

    cout << "TIME (factorization): [ " << LUelapsed[0] << "ms, " 
                                       << LUelapsed[1] << "ms, "
                                       << LUelapsed[2] << "ms]"
         << " (solution): [ " << SOLelapsed[0] << "ms, "
                              << SOLelapsed[1] << "ms, "
                              << SOLelapsed[2] << "ms]"
         << " err: [ " << ERR[0] << ", "
                       << ERR[1] << ", "
                       << ERR[2] << " ]"
         << " cond= " << cond
         << "\n" ;
  }
  return 0;
}
